class EventSchemaInvalidException(Exception):
    pass


class EventNotFoundException(Exception):
    pass


class MappingException(Exception):
    pass


class StrategyNotFoundException(Exception):
    pass
